# -*- encoding: utf-8 -*-
import os
from dna.settings import *
from apep.env_var import get_env_variable, get_env_variable_bool

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "sx)8tn28l=e)wd^v#e!q6w2g3ige!bgskhm-sl%5grv==x$v=("

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

# Application definition
INSTALLED_APPS = [
    "livereload",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "easy_thumbnails",
    "dna",
    "chromosomes",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "livereload.middleware.LiveReloadScript",
]

ROOT_URLCONF = "chromosomes.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

WSGI_APPLICATION = "chromosomes.wsgi.application"

# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases
DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "temp.db"}
}

# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"
    },
]

# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


APP_NAME = "chromosomes"

STATICFILES_DIRS = [os.path.join(APP_NAME, "theme/dist")]

STATIC_ROOT = "static"

STATIC_URL = "/elio_static/"

MEDIA_ROOT = "media"

MEDIA_URL = "/elio_media/"

# Default App Settings
# Because lowercase field names clash with propername Model names!
SILENCED_SYSTEM_CHECKS = [
    "models.E006"
]  # Or add "Model" to the name of each model. Or django-polymorphic?
WEB_PAGE_DNA = {
    "WebPage": {
        "field": [
            "image",
            "name",
            "alternateName",
            "disambiguatingDescription",
            "description",
            "text",
        ],
        "many": ["significantLinks"],
    }
}
DNA_DEPTH = 0
SITE_DNA = {
    "page": {
        APP_NAME.title(): {"engage": WEB_PAGE_DNA},
        "sep1": {"status": "spacer"},
        "Blog": {
            "engage": WEB_PAGE_DNA,
            "page": {
                "Bubbles": {
                    "list": {
                        "CreativeWork": {
                            "paginate": 5,
                            "template": "chromosomes/_list_bubble.html",
                            "field": [
                                "image",
                                "name",
                                "alternateName",
                                "disambiguatingDescription",
                                "description",
                                "text",
                            ],
                        }
                    }
                },
                "Articles": {
                    "list": {
                        "CreativeWork": {
                            "paginate": 10,
                            "field": [
                                "image",
                                "name",
                                "alternateName",
                                "description",
                                "text",
                                "creator",
                                "dateModified",
                                "datePublished",
                                "dateCreated",
                            ],
                            "many": ["comment"],
                        }
                    }
                },
            },
        },
        "Gallery": {
            "engage": WEB_PAGE_DNA,
            "list": {
                "Photograph": {
                    "paginate": 20,
                    "template": "chromosomes/_list_gallery.html",
                    "field": [
                        "name",
                        "alternateName",
                        "description",
                        "image",
                        "creator",
                        "dateCreated",
                    ],
                    "many": ["comment"],
                }
            },
        },
        "About": {
            "engage": {
                "AboutPage": {
                    "field": [
                        "name",
                        "alternateName",
                        "disambiguatingDescription",
                        "description",
                    ],
                    "many": ["reviews"],
                },
                "Organization": {
                    "title": "Contact",
                    "field": ["address", "email", "telephone", "founder"],
                },
            }
        },
    }
}
SCHEMA_VERSION = "3.9"
SCHEMA_PATH = f"schemaorg/data/releases/{SCHEMA_VERSION}/all-layers.jsonld"



from easy_thumbnails.conf import Settings as easy_thumbnails_defaults

THUMBNAIL_WATERMARK = {
    "logo": {"name": "static/apple-touch-icon.png", "opacity": 0.7}
}

THUMBNAIL_PROCESSORS = easy_thumbnails_defaults.THUMBNAIL_PROCESSORS

THUMBNAIL_DEFAULT_OPTIONS = {"size": (100, 100), "crop": "center"}

THUMBNAIL_ALIASES = {
    "": {
        "960x720": {
            "size": (960, 720),
            "crop": False,
            "autocrop": True,
            "background": "white",
            "watermark": "logo",
            "upscale": True,
        },
        "640x480": {
            "size": (640, 480),
            "crop": False,
            "autocrop": True,
            "background": "white",
            "watermark": "logo",
            "upscale": True,
        },
        "300x300": {
            "size": (300, 300),
            "crop": "center",
            "autocrop": True,
            "background": "white",
            "watermark": "logo",
            "upscale": True,
        },
        "100x100": THUMBNAIL_DEFAULT_OPTIONS,
    }
}


# Use CharField with these lengths for these properties.
DNA_SHORT_TEXT_FIELDS = {
    "name": 512,
    "disambiguatingDescription": 512,
    "gtin12": 12,
    "gtin13": 13,
    "gtin14": 14,
    "gtin8": 8,
    "naics": 6,
    "postalCode": 30,
    "flightNumber": 10,
    "branchCode": 10,
    "departureGate": 10,
    "arrivalGate": 10,
    "departureTerminal": 10,
    "arrivalTerminal": 10,
    "iataCode": 3,
    "icaoCode": 4,
    "departurePlatform": 10,
    "arrivalPlatform": 10,
    "busNumber": 10,
    "issn": 8,
    "isbn": 13,
    "isrcCode": 50,
    "iswcCode": 50,
    "vehicleIdentificationNumber": 17,
    "leiCode": 50,
    "courseCode": 50,
    "accessCode": 50,
}

# Use long text for these properties.
DNA_LONG_TEXT_FIELDS = (
    "articleBody",
    "awards",
    "benefits",
    "commentText",
    "dependencies",
    "description",
    "educationRequirements",
    "experienceRequirements",
    "incentives",
    "incentiveCompensation",
    "jobBenefits",
    "lodgingUnitDescription",
    "qualifications",
    "responsibilities",
    "reviewBody",
    "skills",
    "text",
    "transcript",
)

# Because lowercase field names clash with propername Model names!
SILENCED_SYSTEM_CHECKS = [
    "models.E006"
]  # Or add "Model" to the name of each model. Or django-polymorphic?


SCHEMA_VERSION = "3.9"
SCHEMA_PATH = f"schemaorg/data/releases/{SCHEMA_VERSION}/all-layers.jsonld"



# -*- encoding: utf-8 -*-
from django.urls import reverse_lazy
from slugify.slugify import slugify


def pages_deeply(pages_dna):
    """Flatten out all the pages in the SITE_DNA for easy searching."""

    def items():
        for key, value in pages_dna.get("page", {}).items():
            yield key, value
            if isinstance(value, dict):
                for subkey, subvalue in pages_deeply(value).items():
                    yield subkey, subvalue

    return dict(items())


def get_page_dna(site_dna, current_page):
    """Find the details of the current page in the SITE_DNA."""
    all_pages = pages_deeply(site_dna)
    return all_pages.get(current_page, {})


def menu_lookahead(sub_pages_dna, current_page):
    """A lookahead to return if a menu has sub menus activated."""
    all_sub_pages = pages_deeply(sub_pages_dna)
    return current_page in all_sub_pages.keys()


def get_page_name_from_slug(site_dna, page_slug, thing_name):
    """Find the details of the current page in the SITE_DNA."""
    all_pages = pages_deeply(site_dna)
    for menu_key in all_pages.keys():
        if slugify(menu_key) == page_slug:
            return menu_key
    # If you can't find the page_key then it's probably an isolated thing.
    return thing_name


def get_model_dna(page_dna, thing_name):
    """Find the details of a model of the current page in the SITE_DNA."""
    if not isinstance(page_dna, dict):
        return {}
    for elio_type in ["engage", "list"]:
        if thing_name in list(page_dna.get(elio_type, {})):
            if isinstance(page_dna[elio_type], dict):
                return page_dna[elio_type].get(thing_name, {})
    return {}


def get_page_menu(site_dna, current_page, site_things=[]):
    """Convert the SITE_DNA into a menu."""
    menu = dict()
    # Start with every key of the current "page" dict.
    for page_name, page_vals in site_dna.get("page", {}).items():
        # Start building the menu one menuitem at a time.
        menu[page_name] = dict()
        # Start give this item a URL
        menu[page_name]["url"] = reverse_lazy(
            "engage.page", kwargs={"page_slug": slugify(page_name)}
        )
        # Give the menuitem any status set in the SITE_DNA
        if page_vals.get("status"):
            menu[page_name]["status"] = page_vals.get("status")

        # Change the status to active if this menuitem is current page.
        if slugify(page_name) == slugify(current_page):
            menu[page_name]["status"] = "active"

        # If page is active, or (looking ahead) a sub page is active....
        if menu[page_name].get("status") or menu_lookahead(
            page_vals, current_page
        ):
            # If it was a subitem that was active (looking ahead).
            if not menu[page_name].get("status"):
                # Flag as a crumb to the active menu.
                menu[page_name]["status"] = "crumb"  # Make the breadcrumbs
            # Build the sub menu for this menuitem if active.
            sub_menu = get_page_menu(page_vals, current_page)
            # Add the sub to the current menuitem.
            if sub_menu.get("page"):
                menu[page_name].update(sub_menu)
    # Special case: Add a Browse Things section to the menu for all Things.
    if site_things:
        menu["Browse Things"] = {"status": "label", "page": {}}
        for thing in sorted(site_things):
            menu["Browse Things"]["page"][thing] = dict()
            menu["Browse Things"]["page"][thing]["url"] = reverse_lazy(
                "list.things", kwargs={"page_slug": thing, "thing": thing}
            )
            if thing == current_page:
                menu["Browse Things"]["page"][thing]["status"] = "active"
    return {"page": menu}



# -*- encoding: utf-8 -*-
import copy
from django.apps import apps
from django.conf import settings
from django.core.paginator import Paginator
from django.urls import reverse_lazy
from slugify.slugify import slugify
from dna.dna import DnaOfDjangoModel
from dna.forms import every_form
from dna.view_utils import (
    get_model_dna,
    get_page_dna,
    get_page_menu,
    get_page_name_from_slug,
    menu_lookahead,
)


class EveryThingMixin:
    _dna = None

    @property
    def site_dna(self):
        return copy.deepcopy(settings.SITE_DNA)

    @property
    def dna(self):
        if not self._dna:
            self._dna = DnaOfDjangoModel(
                settings.SCHEMA_PATH, self.site_dna, settings.DNA_DEPTH
            )
            self._dna.build()
        return self._dna

    @property
    def page_slug(self):
        return self.kwargs.get("page_slug", "")

    @property
    def page_name(self):
        return get_page_name_from_slug(
            self.site_dna, self.page_slug, self.thing_name
        )

    @property
    def thing_name(self):
        return self.kwargs.get("thing", "")

    @property
    def page_link(self):
        if self.page_name == self.thing_name:
            return reverse_lazy(
                "list.things",
                kwargs={"page_slug": self.page_slug, "thing": self.thing_name},
            )
        else:
            return reverse_lazy(
                "engage.page", kwargs={"page_slug": self.page_slug}
            )

    @property
    def page_menu(self):
        return get_page_menu(self.site_dna, self.page_name, self.dna.backbone)

    @property
    def Model(self):
        if self.thing_name:
            return apps.get_model(
                model_name=self.thing_name, app_label=settings.APP_NAME.lower()
            )
        return None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["app_name"] = settings.APP_NAME.title()
        context["home_page"] = list(self.page_menu.get("page", {}))[0]
        context["page_name"] = self.page_name
        context["page_slug"] = self.page_slug
        context["thing_name"] = self.thing_name
        context["page_menu"] = self.page_menu
        return context


class EveryQuerySetMixin:

    paginate_by = 20

    def get_queryset(self):
        return self.Model.objects.all().order_by("name")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Don't extend the "page" querystring for standard lists.
        context["paginate_type"] = ""
        return context


class EveryGetObjectMixin:
    def get_object(self):
        obj = self.Model.objects.get(pk=self.kwargs.get("pk"))
        return obj


class EveryEditMixin:
    def get_form_class(self):
        page_dna = get_page_dna(self.site_dna, self.page_name)
        model_dna = get_model_dna(page_dna, self.thing_name)
        elio_key = slugify(f"{self.page_name}-{self.thing_name}")
        field = model_dna.get("field", [])
        return every_form(
            self.Model, elio_key, self.page_link, model_dna.get("field", [])
        )


class EveryDeleteMixin:
    def get_success_url(self):
        return self.page_link


class EveryPageMixin:
    def get_template_names(self):
        """Decide whether to use the default template, or a customer template
        specified in the `SITE_DNA`."""
        custom_template = get_page_dna(self.site_dna, self.page_name).get(
            "template"
        )
        return custom_template or self.template_name

    def get_page_content(self, elio_type, default_template):
        """Build a content block.
        - An "engage" block is treated like a singleton for this page and types.
        - A "list" block handles a queryset for this page and types."""
        # Find out what schema will drive the content for this page.
        content = []
        page_dna = get_page_dna(self.site_dna, self.page_name).get(elio_type)
        # Does the SITE_DNA expect content of this elio_type?
        if page_dna:
            # For each thing type listed as set/dict/list.
            for thing_name in list(page_dna):
                elio_key = slugify(f"{self.page_name}-{thing_name}")
                # Get the model for this thing type.
                ContentThing = apps.get_model(
                    model_name=thing_name,  # settings.APP_NAME.title() + thing_name ,
                    app_label=settings.APP_NAME.lower(),
                )
                # Get the content as a singleton or queryset as per elio_type.
                if elio_type == "engage":
                    try:
                        # Get the record for this page and thing type.
                        content_thing = ContentThing.objects.get(
                            elio_key=elio_key,
                            elio_page=self.page_link,
                            elio_role=elio_type,
                        )
                    except ContentThing.DoesNotExist:
                        # Build a unique record for this page and thing type.
                        content_thing = ContentThing(
                            name=f"{self.page_name} {thing_name}",
                            elio_key=elio_key,
                            elio_page=self.page_link,
                            elio_role=elio_type,
                        )
                        content_thing.save()
                else:
                    content_thing = ContentThing.objects.filter(
                        elio_key=elio_key,
                        elio_page=self.page_link,
                        elio_role=elio_type,
                    ).order_by("name")
                # Build a dict for this content.
                thing_dna = {}
                if isinstance(page_dna, dict):
                    # The SITE_DNA may have settings like pagination or field.
                    thing_dna = page_dna.get(thing_name, {})
                thing_dna[elio_type] = content_thing
                thing_dna["thing_name"] = thing_name
                # Add a paginator if required.
                if thing_dna.get("paginate") and elio_type == "list":
                    paginator = Paginator(
                        content_thing, thing_dna.get("paginate")
                    )
                    thing_dna[elio_type] = paginator.get_page(
                        self.request.GET.get("page" + thing_name)
                    )
                    thing_dna["paginator"] = paginator
                    thing_dna["is_paginated"] = True
                    thing_dna["paginate_type"] = thing_name
                else:
                    thing_dna["paginator"] = None
                    thing_dna["is_paginated"] = False
                    thing_dna["paginate_type"] = ""
                # Always have a template.
                thing_dna["template"] = (
                    thing_dna.get("template") or default_template
                )
                # Append it to the content for this page and elio_type.
                content.append(thing_dna)
        return content

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["page_engages"] = self.get_page_content(
            "engage", default_template="dna/_thing_engaged.html"
        )
        context["page_lists"] = self.get_page_content(
            "list", default_template="dna/_thing_listed.html"
        )
        return context



# -*- encoding: utf-8 -*-
import os
from dna.settings import *
from apep.env_var import get_env_variable, get_env_variable_bool

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "sx)8tn28l=e)wd^v#e!q6w2g3ige!bgskhm-sl%5grv==x$v=("

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

# Application definition
INSTALLED_APPS = [
    "livereload",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "easy_thumbnails",
    "dna",
    "molecule",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "livereload.middleware.LiveReloadScript",
]

ROOT_URLCONF = "molecule.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

WSGI_APPLICATION = "molecule.wsgi.application"

# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases
DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "temp.db"}
}

# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"
    },
]

# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


APP_NAME = "molecule"

STATICFILES_DIRS = [os.path.join(APP_NAME, "theme/dist")]

STATIC_ROOT = "static"

STATIC_URL = "/static/"

MEDIA_ROOT = "media"

MEDIA_URL = "/media/"

# Default App Settings
# Because lowercase field names clash with propername Model names!
SILENCED_SYSTEM_CHECKS = [
    "models.E006"
]  # Or add "Model" to the name of each model. Or django-polymorphic?
SITE_DNA = {
    "page": {
        "A Grand Idea": {"engage": {"Thing": {"field": ["name"]}}},
        "Universal": {
            "template": "molecule/universal.html",
            "engage": ["Universe"],
            "page": {
                "Galaxies": {
                    "engage": {"Thing": {"field": []}},
                    "list": {
                        "Galaxy": {
                            "field": ["name"],
                            "paginate": 3,
                            "template": "molecule/_galaxy_listed.html",
                        }
                    },
                    "page": {
                        "MilkyWay": {
                            "list": ["SolarSystem", "Sun"],
                            "page": {
                                "Sol": {
                                    "list": {
                                        "Moon": {
                                            "field": ["name"],
                                            "paginate": 2,
                                        },
                                        "Planet": {
                                            "field": ["name"],
                                            "paginate": 2,
                                            "list": ["Moon", "Satellite"],
                                        },
                                    }
                                }
                            },
                        }
                    },
                }
            },
        },
        "Roaming": {"list": ["Moon", "Satellite", "GPS", "Asteroid"]},
    }
}
DNA_DEPTH = 1
SCHEMA_VERSION = "3.9"
SCHEMA_PATH = "molecule/tests/mocks/space.jsonld"


# Use CharField with these lengths for these properties.
DNA_SHORT_TEXT_FIELDS = {
    "name": 512,
    "disambiguatingDescription": 512,
    "gtin12": 12,
    "gtin13": 13,
    "gtin14": 14,
    "gtin8": 8,
    "naics": 6,
    "postalCode": 30,
    "flightNumber": 10,
    "branchCode": 10,
    "departureGate": 10,
    "arrivalGate": 10,
    "departureTerminal": 10,
    "arrivalTerminal": 10,
    "iataCode": 3,
    "icaoCode": 4,
    "departurePlatform": 10,
    "arrivalPlatform": 10,
    "busNumber": 10,
    "issn": 8,
    "isbn": 13,
    "isrcCode": 50,
    "iswcCode": 50,
    "vehicleIdentificationNumber": 17,
    "leiCode": 50,
    "courseCode": 50,
    "accessCode": 50,
}

# Use long text for these properties.
DNA_LONG_TEXT_FIELDS = (
    "articleBody",
    "awards",
    "benefits",
    "commentText",
    "dependencies",
    "description",
    "educationRequirements",
    "experienceRequirements",
    "incentives",
    "incentiveCompensation",
    "jobBenefits",
    "lodgingUnitDescription",
    "qualifications",
    "responsibilities",
    "reviewBody",
    "skills",
    "text",
    "transcript",
)



"""DnaOfDjangoModel: Return the DNA of django models.

This class builds a python dict we can use to replicate Django models.

The dictionary is a cacheable/pickle friendly object which can be passed to the
`ReplicateDjangoModel` methods. Having this in another layer allows us to store
it and also peek at it which is helpful for debugging.

**Usage**

::

    from dna.dna import DnaOfDjangoModel

    dna = DnaOfDjangoModel("schema.jsonld", site_dna, type_depth)
    replicate_dna = dna.build()"""
# -*- encoding: utf-8 -*
import hashlib
import json
from django.apps import apps
from django.db import models
from django.conf import settings
from django.urls import reverse
from dna.crispr import Crispr
from apep.picklejar import PickleJar


class DnaOfDjangoModel:
    """Build the DNA of django models from the Crispr."""

    def __init__(
        self,
        schema_path,
        site_dna,
        type_depth,
        domain="http://schema.org",
        no_comments=False,
    ):
        # How deep to use class types
        self.type_depth = type_depth
        # A map of the site
        self.site_dna = site_dna
        # Convert schema.org jsonld information into parsable units.
        self.crispr = Crispr(schema_path, domain=domain)
        # Remove comments from the build.
        self.no_comments = no_comments
        # Holding places for the structure `DnaOfDjangoModel` builds.
        self._replicate_dna = dict()
        self._backbone = list()
        self._iterables = set()
        self._site_things = dict()
        self._thing_names = set()

    @property
    def domain(self):
        """The schema's domain."""
        return self.crispr.domain

    @property
    def replicate_dna(self):
        return self._replicate_dna

    @property
    def backbone(self):
        return [self.crispr.name_from_url(b) for b in self._backbone]

    @property
    def iterables(self):
        return self._iterables

    @property
    def site_things(self):
        return self._site_things

    @staticmethod
    def dna_of_model_field_type_selector(first_type_in, this_list):
        """Gets the first type which matches the list."""
        for type in first_type_in:
            if type in this_list:
                return type
        return None

    def dna_of_model_field_type(self, thing_name, prp):
        """Selects the best django field Type for any property.

        This works on a sliding scale: Taking the first Type which matches if
        the Property has many.

        1. Use the property_name against predefined lists.
        2. Use the Property Enumerated if Enumerated Type is in dependencies.
        3. Use a Class Type if the Class Type is already in dependencies.
        4. Use the first Django Field for Primitive or semi-Primitive.
        5. Use a DJANGO_FIELDS for Primitive Types.
        6. Use Text by default.
        """
        property_name = Crispr.property_name(prp)
        IMAGEFIELD = dict(
            blank=True,
            field_type="ImageField",
            upload_to=f"images/{thing_name}/{property_name}",
            null=True,
        )
        FILEFIELD = dict(
            blank=True,
            field_type="FileField",
            upload_to=f"files/{thing_name}/{property_name}",
            null=True,
        )
        MONEYFIELD = dict(
            decimal_places=2, field_type="DecimalField", max_digits=10
        )

        # Beware 1. The order here is important. Types are selected first come.
        # Beware 2. No field should *python-by-reference* the same Type.
        DJANGO_FIELDS = {
            "Boolean": dict(
                blank=False,
                default=False,
                field_type="BooleanField",
                null=False,
            ),
            "Time": dict(blank=True, field_type="TimeField", null=True),
            "DateTime": dict(blank=True, field_type="DateTimeField", null=True),
            "Date": dict(blank=True, field_type="DateField", null=True),
            "Number": dict(
                blank=True, default=0, field_type="FloatField", null=True
            ),
            "Integer": dict(
                blank=True, default=0, field_type="IntegerField", null=True
            ),
            "Float": dict(
                decimal_places=10, field_type="DecimalField", max_digits=19
            ),
            "Quantity": dict(
                blank=True, default=0, field_type="FloatField", null=True
            ),
            "Duration": dict(blank=True, field_type="DurationField", null=True),
            "URL": dict(
                blank=True, field_type="URLField", max_length=512, null=True
            ),
            "email": dict(field_type="EmailField"),
            "image": IMAGEFIELD,
            "logo": IMAGEFIELD,
            "screenshot": IMAGEFIELD,
            "beforeMedia": FILEFIELD,
            "afterMedia": FILEFIELD,
            "duringMedia": FILEFIELD,
            "minPrice": MONEYFIELD,
            "maxPrice": MONEYFIELD,
            "price": MONEYFIELD,
            "minValue": MONEYFIELD,
            "value": MONEYFIELD,
            "maxValue": MONEYFIELD,
        }
        property_types = self.crispr.property_types_of(prp, self.domain)
        type = None  # Currently unknown.
        # Is there a matching Class Type?
        class_type = DnaOfDjangoModel.dna_of_model_field_type_selector(
            [t for t in property_types if not t.endswith("/Text")],
            [t for t in self._backbone if t not in self.crispr.primitives],
        )
        enumerated_type = self.crispr.property_enumerated(prp)
        choices = self.crispr.enumerations_of_by_url(enumerated_type)
        if enumerated_type and class_type and choices:
            # Type Enumerated
            return dict(
                default=choices[0][0],
                choices=choices,
                field_type="CharField",
                help_text="" if self.no_comments else prp["rdfs:comment"],
                max_length=len(max([c[0] for c in choices], key=len)),
            )
        if class_type:
            # Use the Class Type
            type = class_type
        if not type:
            # Try a Django Model
            type = DnaOfDjangoModel.dna_of_model_field_type_selector(
                [f"{self.domain}/{t}" for t in DJANGO_FIELDS.keys()],
                [t for t in property_types if not "Text" in t],
            )
        if not type:
            # Try a Primitive Type - leave Text til last
            type = DnaOfDjangoModel.dna_of_model_field_type_selector(
                [t for t in property_types if not "Text" in t], self._backbone
            )
        if not type:
            # Fall back to Text
            type = f"{self.domain}/Text"
        # Get the name
        type_name = self.crispr.name_from_url(type)
        # Now decide on the best django model Type
        django_type = None
        # Name of the property prefect for a known django model Type.
        if property_name in DJANGO_FIELDS.keys():
            django_type = DJANGO_FIELDS[property_name]
        # Name of the property preordained for long text.
        elif property_name in settings.DNA_LONG_TEXT_FIELDS:
            django_type = dict(blank=True, field_type="TextField")
        # Name of the property preordained char length.
        elif property_name in settings.DNA_SHORT_TEXT_FIELDS.keys():
            django_type = dict(
                blank=True,
                field_type="CharField",
                max_length=settings.DNA_SHORT_TEXT_FIELDS[property_name],
            )
        # Type matches a known django model Type.
        elif type_name in DJANGO_FIELDS.keys():
            django_type = DJANGO_FIELDS[type_name]
        # Standard Text Type.
        elif type_name == "Text":
            django_type = dict(
                blank=True, field_type="CharField", max_length=255
            )
        # ForeignKey to another schema Type.
        else:
            django_type = dict(
                blank=True,
                field_type="ForeignKey",
                null=True,
                related_name=f"{settings.APP_NAME}_{type_name}_{thing_name}_{property_name}".lower(),
                related_to=type_name,  # f"{settings.APP_NAME.title()}{type_name}",
            )
        django_type["help_text"] = (
            "" if self.no_comments else prp["rdfs:comment"]
        )
        return django_type

    def dna_of_model_fields(self, thing_name, thing_properties):
        """Return the django fields converted from the properties of a model.

        :param thing_name: e.g. "AppnameThing"
        :param thing_properties: e.g.
            [{
                '@id': 'http://schema.org/sameAs',
                '@type': 'rdf:Property',
                'http://schema.org/domainIncludes': {'@id': 'http://schema.org/Thing'},
                'http://schema.org/rangeIncludes': {'@id': 'http://schema.org/URL'},
                'rdfs:comment': "Comment...",
                'rdfs:label': 'sameAs'
            }, {
                '@id': 'http://schema.org/url',
                '@type': 'rdf:Property',
                'http://schema.org/domainIncludes': {'@id': 'http://schema.org/Thing'},
                'http://schema.org/rangeIncludes': {'@id': 'http://schema.org/URL'},
                'rdfs:comment': 'URL of the item.',
                'rdfs:label': 'url'
            }, ...]
        """
        fields = dict()
        for prp in thing_properties:
            fields[Crispr.property_name(prp)] = self.dna_of_model_field_type(
                thing_name, prp
            )
        return fields

    def model_inherits_fields(self, immediately_inherits_from):
        """All fields a model will inherit given its immediate inheritance."""
        inherits_fields = []
        for inheritance in immediately_inherits_from:
            if not inheritance == "models.Model":
                inheritance_dna = self._replicate_dna[inheritance]
                inherits_fields += inheritance_dna["fields_of"].keys()
                inherits_fields += self.model_inherits_fields(
                    inheritance_dna["inherits_from"]
                )
        return inherits_fields

    def model_inherits_from(self, class_of_thing):
        """Return which django model this sub class should inherit from.

        :param class_of_thing: e.g.
            {
              "@id": "http://schema.org/CafeOrCoffeeShop",
              "@type": "rdfs:Class",
              "rdfs:comment": "A cafe or coffee shop.",
              "rdfs:label": "CafeOrCoffeeShop",
              "rdfs:subClassOf": {
                "@id": "http://schema.org/FoodEstablishment"
              }
            }
        """
        # Assume this thing inherits from the base Model (only Thing does!)
        inherits_from = ["models.Model"]
        subclasses_things = Crispr.subclasses_direct(class_of_thing)
        if subclasses_things:
            inherits_from = []
            for subclasses in subclasses_things:
                model_dna = self.dna_of_model(
                    self.crispr.class_of_by_url(subclasses)
                )
                inherits_from.append(model_dna["thing_name"])
        return inherits_from

    def dna_of_model_exists(self, thing_name):
        """Utility to determine if the thing_name exists in our structure."""
        return thing_name in [m for m in self._replicate_dna.keys()]

    def dna_of_model_if_exists(self, thing_name):
        """Utility to get an existing django model if has already been built."""
        model = None
        if self.dna_of_model_exists(thing_name):
            model = self._replicate_dna[thing_name]
        return model

    def dna_of_model(self, class_of_thing, options=None):
        """Return a django model given a class object.

        :param class_of_thing: e.g.
            {
                '@id': 'http://schema.org/Thing',
                '@type': 'rdfs:Class',
                'rdfs:comment': 'The most generic type of item.',
                'rdfs:label': 'Thing'
            }
        """
        thing_name = class_of_thing.get("rdfs:label")
        if isinstance(thing_name, dict):
            thing_name = thing_name.get("@value")
        # thing_name = f"{settings.APP_NAME.title()}{thing_name}"
        # Simply return dna of models we have already created (i.e. inherit from).
        model = self.dna_of_model_if_exists(thing_name)
        if model:
            return model
        # Fields
        fields_of = self.dna_of_model_fields(
            thing_name, self.crispr.properties_of_by_url(class_of_thing["@id"])
        )
        # Establish all the inheritance from other things
        inherits_from = self.model_inherits_from(class_of_thing)
        # Remove fields inherited from parent classes (it's possible because in
        # schema a child class inherits from 2 classes - and any of them
        # are optional - so schema shows properties in the child class that
        # should be present even if you opt not to inherit one of the parents
        # with it).
        all_parent_fields = set(self.model_inherits_fields(inherits_from))
        names_existing = all_parent_fields.intersection(set(fields_of))
        if names_existing:
            for field_name in names_existing:
                fields_of.pop(field_name)
        # Switch _iterables into Many2Many
        iterable_fields = dict()
        for field_name, field_dna in fields_of.items():

            if field_name in self._iterables:
                iterable_fields[field_name] = field_dna.copy()
                if field_dna["field_type"] == "ForeignKey":
                    del iterable_fields[field_name]["blank"]
                    del iterable_fields[field_name]["null"]
                    del iterable_fields[field_name]["related_name"]
                else:
                    iterable_fields[field_name] = dict(
                        field_type="ForeignKey",
                        related_to="Thing",  # f"{settings.APP_NAME.title()}Thing",
                    )
        # Remove the iterable_fields from
        #       TODO: Consider not removing the field. It won't appear.
        if iterable_fields:
            for field_name in iterable_fields.keys():
                del fields_of[field_name]
        # Build the dict
        model_dict = {
            "fields_of": fields_of,
            "inherits_from": inherits_from,
            "thing_name": thing_name,
            "iterables": iterable_fields,
        }
        # Keep track of the dna of models and fields we have created.
        self._replicate_dna[thing_name] = model_dict
        return model_dict

    def _get_things_of_site(self, site_dna, thing_names=None):
        """Recursively extract all the model Thing Types from the SITE_DNA."""
        if not thing_names:
            thing_names = {"Thing"}
        for key, value in site_dna.items():
            if isinstance(value, dict):
                # Fetch "Things Types" hosted in "engage" and "list" keys.
                for elio_type in ["engage", "list"]:
                    for each_thing in list(value.get(elio_type, {})):
                        self._site_things[key] = {
                            each_thing: value.get(elio_type, {}).copy()
                        }
                        thing_names.add(each_thing)
                # Thing properties hosted in the "iterate" key should be listed
                # as Many2Many objects.
                iterate_properties = value.get("iterate", [])
                for each_property in iterate_properties:
                    self._iterables.add(each_property)
                    prop = self.crispr.property_of_by_name(each_property)
                    # Look for non-primitive "Thing Types".
                    prop_types = Crispr.property_types_of(prop, self.domain)
                    prim_types = self.crispr.primitive_types_of(prop_types)
                    # "Thing Types" when more Types than just Primitive.
                    thing_types = set(prop_types).difference(set(prim_types))
                    for thing_type in thing_types:
                        thing_names.add(self.crispr.name_from_url(thing_type))
                # Recusively look for more Types.
                thing_names.union(self._get_things_of_site(value, thing_names))
        return thing_names

    def _get_replicate_dna(self, backbone):
        """Build all the DNA of models listed in the backbone."""
        # Do each model in any order: `dna_of_model` will do ancestors first.
        for thing_url in self._backbone:
            class_of = self.crispr.class_of_by_url(thing_url)
            self.dna_of_model(class_of)
        return self._replicate_dna

    def build(self):
        """Main build method."""
        site_dna_dump = json.dumps(self.site_dna, sort_keys=True)
        picklename = hashlib.sha1(
            f"{site_dna_dump}{self.type_depth}".encode()
        ).hexdigest()
        pick = PickleJar(f"picklings/{settings.APP_NAME}", picklename[:200])
        # Return an existing DNA.
        if pick.ripe:
            dna_dna = pick.open()
            self._replicate_dna = dict(dna_dna["_replicate_dna"])
            self._backbone = list(dna_dna["_backbone"])
            self._iterables = set(dna_dna["_iterables"])
            self._site_things = dict(dna_dna["_site_things"])
            self._thing_names = list(dna_dna["_thing_names"])
            return self._replicate_dna
        # Get the list of all things the site_dna models mention.
        self._thing_names = self._get_things_of_site(self.site_dna)
        # Get the list of all things the site_dna models require.
        self._backbone = self.crispr.dependencies_of_by_names(
            list(self._thing_names), self.type_depth
        )
        # Get the DNA we can pass to the Replicate class
        self._replicate_dna = self._get_replicate_dna(self._backbone)
        # Pickle for next time
        pick.pickle(
            {
                "_replicate_dna": self._replicate_dna,
                "_site_things": self._site_things,
                "_iterables": list(self._iterables),
                "_thing_names": list(self._thing_names),
                "_backbone": self._backbone,
            }
        )
        return self._replicate_dna



# -*- encoding: utf-8 -*-
from django import template
from dna.watermark import watermark


register = template.Library()


@register.simple_tag(name="thumbnail")
def thumbnail(image_path, dimensions):
    """Marshall call to the `watermark` method and return an easy_thumbnail."""
    thumb = watermark(image_path, dimensions)
    return thumb


@register.filter("has_no_merit")
def has_no_merit(value):
    """Determine if a field value has any merit at all."""
    if value:
        return not bool(str(value).strip())
    return (
        not value
        or value == "False"
        or value is None
        or value == "None"
        or str(value) == "0"
        or str(value) == "0.0"
    )


@register.filter("is_in")
def is_in(value, comma_list):
    """See if a value is in a comma separated list."""
    return value in comma_list.split(",")


# Development only.
@register.filter("dirt")
def dirt(value):
    """'Dish the dirt' on a template variable render the output of `dir`."""
    return dir(value)


@register.filter("flip_boo")
def flip_boo(value):
    """Flip a boolean value from True to False and back."""
    return bool(not value)


# need to add 'django.core.context_processors.request' to settings.TEMPLATES
@register.simple_tag(takes_context=True)
def qrystr_in_concert(context, name, value):
    """Replace a querystring parameter in concert with existing querystrings."""
    request = context["request"]
    dict_ = request.GET.copy()
    dict_[name] = value
    return dict_.urlencode()



# -*- coding: utf-8 -*-
import os
import setuptools
from distutils.core import setup


def read_file_into_string(filename):
    path = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(path, filename)
    try:
        return open(filepath).read()
    except IOError:
        return ""


def get_readme():
    if os.path.exists("README.md"):
        return read_file_into_string("README.md")
    return ""


setup(
    name="elio-dna",
    packages=[
        "dna",
        "dna.management",
        "dna.management.commands",
        "dna.templatetags",
    ],
    package_data={"dna": ["templates/*.*", "templates/dna/*.*"]},
    install_requires=[
        "elio-apep",
        "Django",
        "django-polymorphic",
        "easy-thumbnails",
        "Pillow",
        "python-slugify",
    ],
    include_package_data=True,
    version="1.0.4",
    description="Django models, views, forms, and templates, built on schema.org Things the elioWay.",
    author="Tim Bushell",
    author_email="tcbushell@gmail.com",
    url="https://elioway.gitlab.io/eliothing/dna",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Console",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System",
    ],
    license="MIT",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
)



# -*- encoding: utf-8 -*-
from django.urls import reverse_lazy
from django.views.generic import DetailView, ListView
from django.views.generic.base import TemplateView
from django.views.generic.edit import CreateView, DeleteView, UpdateView
from dna.view_mixins import (
    EveryThingMixin,
    EveryQuerySetMixin,
    EveryGetObjectMixin,
    EveryEditMixin,
    EveryPageMixin,
    EveryDeleteMixin,
)


class EveryCreateViewMixin(
    EveryThingMixin, EveryGetObjectMixin, EveryEditMixin, CreateView
):
    pass


class EveryUpdateViewMixin(
    EveryThingMixin, EveryGetObjectMixin, EveryEditMixin, UpdateView
):
    pass


class EveryDeleteViewMixin(
    EveryThingMixin,
    EveryGetObjectMixin,
    EveryQuerySetMixin,
    EveryDeleteMixin,
    DeleteView,
):
    pass


class EveryEngageViewMixin(
    EveryThingMixin, EveryQuerySetMixin, EveryGetObjectMixin, DetailView
):
    pass


class EveryIterateViewMixin(EveryThingMixin, TemplateView):
    pass


class EveryOptimizeViewMixin(EveryThingMixin, TemplateView):
    pass


class EveryListViewMixin(EveryThingMixin, EveryQuerySetMixin, ListView):
    pass


class EveryPageViewMixin(EveryThingMixin, EveryPageMixin, TemplateView):
    pass


class EveryHomeViewMixin(EveryThingMixin, TemplateView):
    pass
